/* @(#)scsi-linux-sg.c  1.30 98/11/06 Copyright 1997 J. Schilling */
#ifndef lint
/* sccs string removed because this version was produced with a "read-only"
   version 1.30 as a starting point - dpg */
#endif
/*
 *  Interface for Linux generic SCSI implementation (sg).
 *
 *  This is the interface for the broken Linux SCSI generic driver.
 *  This is a hack, that tries to emulate the functionality
 *  of the scg driver.
 *
 *  Design flaws of the sg driver:
 *  -   cannot see if SCSI command could not be send
 *  -   cannot get SCSI status byte
 *  -   cannot get real dma count of tranfer
 *  -   cannot get number of bytes valid in auto sense data
 *  -   to few data in auto sense (CCS/SCSI-2/SCSI-3 needs >= 18)
 *
 *  This code contains support for the sg driver version 2
 *
 *  Copyright (c) 1997 J. Schilling
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * As per the above "GPL" this code has been modified by
 * Douglas Gilbert (abbreviated to "dpg" in code comments) 990328 
 * Please report problems to dgilbert@interlog.com . 
 *
 * This code is an attempt to cope "seamlessly" with the various versions
 * of the Linux sg at this time. Runtime identification of the sg version
 * via ioctls is attempted. Three versions are identified:
 *      - original Linux sg device driver (kernels 2.0.x, 2.2.0->4)
 *      - Heiko Eissfeldt's Nov '98 sg mods
 *      - D. Gilbert's sg mods (versions 2.1.6->9, 2.1.30+)
 * The following concerns have also been addressed:
 *      - scanning failures on 2nd + subsequent adapters
 *      - correctly register device ids >= 8 (for wide scsi buses)
 *      - apply ioctls to all fds rather than the first one and assume the
 *        same result for other fds
 *      - scsi_getbuf() yields a buffer size == min(SG_GET_RESERVED_SIZE
 *        for all scanned devices). 0 length buffer possible => don't burn
 *      - open sg devices O_NONBLOCK and skip devices that are EBUSY
 *        (rather than hang on them as it has with cdparanoia around)
 *        and later switch them to blocking for data transfer
 */

#include <linux/version.h>

#ifndef LINUX_VERSION_CODE  /* Very old kernel? */
#   define LINUX_VERSION_CODE 0
#endif

#if LINUX_VERSION_CODE >= 0x01031a /* <linux/scsi.h> introduced in 1.3.26 */
#if LINUX_VERSION_CODE >= 0x020000 /* <scsi/scsi.h> introduced somewhere. */
/* Need to fine tune the ifdef so we get the transition point right. */
#include <scsi/scsi.h>
#else
#include <linux/scsi.h>
#endif
#else
#define __KERNEL__
#include <linux/fs.h>
#undef __KERNEL__
#include "block/blk.h"
#include "scsi/scsi.h"
#endif

#include "scsi/sg.h"


/* #define DPG_DEBUG */

#ifndef SCSI_IOCTL_GET_BUS_NUMBER
#define SCSI_IOCTL_GET_BUS_NUMBER 0x5386
#endif

#ifndef SG_GET_BUFSIZE
#define SG_GET_BUFSIZE 0x2206
#endif

#ifndef SG_GET_RESERVED_SIZE
#define SG_GET_RESERVED_SIZE 0x2272
#endif

#define SG_SCSI_CHECK_CONDITION 2   /* N.B. in Linux CHECK_CONDITION==1 */

#ifndef sg_cmd_status
#define sg_cmd_status pack_len
#endif

#ifndef SG_MAX_SENSE
#   define  SG_MAX_SENSE    16  /* Too small for CCS / SCSI-2   */
#endif                  /* But cannot be changed    */

/* Which sg interface */
#define SG_INTF_UNKN 0  /* Unknown at this point */
#define SG_INTF_ORIG 1
#define SG_INTF_NV 2    /* Interface defined by cdrecord 1.8 in Nov '98 */
#define SG_INTF_JA 3    /* Interface of sg in 2.2.4-ac1 (v2.1.30 or later) */

static int sg_intf = SG_INTF_UNKN;

typedef struct sg_header_nv {
    int pack_len;       /* length of incoming packet (including header) / */
    int reply_len;      /* maximum length of expected reply / actual transfer count */
    int pack_id;                  /* id number of packet */
    int result;                   /* 0==ok, otherwise refer to errno codes */
    unsigned int twelve_byte :1;  /* Force 12 byte command length (unused if want_new is set) */
    unsigned int want_new    :1;  /* User requests new behaviour */
    unsigned int grant_new   :1;  /* Driver grants new behaviour */
    unsigned int cdb_len     :5;  /* Command descriptor block length 6..31 (mandatory) */
    unsigned int sense_len   :5;  /* Set max / get actual sense length 0..31 (mandatory) */
    unsigned int other_flags :19; /* for future use */
    unsigned char sense_buffer[SG_MAX_SENSE];/* used only by reads */
} Sgh_nv;

typedef struct sg_header_ja {
    int pack_len;    /* [o] reply_len (ie useless), ignored as input */
    int reply_len;   /* [i] max length of expected reply (inc. sg_header) */
    int pack_id;     /* [io] id number of packet (use ints >= 0) */
    int result;      /* [o] 0==ok, else (+ve) Unix errno code (e.g. EIO) */
    unsigned int twelve_byte:1;
        /* [i] Force 12 byte command length for group 6 & 7 commands  */
    unsigned int target_status:5;   /* [o] scsi status from target */
    unsigned int host_status:8;     /* [o] host status (see "DID" codes) */
    unsigned int driver_status:8;   /* [o] driver status+suggestion */
    unsigned int other_flags:10;    /* unused */
    unsigned char sense_buffer[SG_MAX_SENSE]; /* [o] when target_status is
           CHECK_CONDITION or COMMAND_TERMINATED [or ...] this is output. */
} Sgh_ja;

/*
 * XXX There must be a better way than duplicating things from system include
 * XXX files. This is stolen from /usr/src/linux/drivers/scsi/scsi.h
 */
#ifndef DID_OK
#define DID_OK          0x00 /* NO error                                */
#define DID_NO_CONNECT  0x01 /* Couldn't connect before timeout period  */
#define DID_BUS_BUSY    0x02 /* BUS stayed busy through time out period */
#define DID_TIME_OUT    0x03 /* TIMED OUT for other reason              */
#define DID_BAD_TARGET  0x04 /* BAD target.                             */
#define DID_ABORT       0x05 /* Told to abort for some other reason     */
#define DID_PARITY      0x06 /* Parity error                            */
#define DID_ERROR       0x07 /* Internal error                          */
#define DID_RESET       0x08 /* Reset by somebody.                      */
#define DID_BAD_INTR    0x09 /* Got an interrupt we weren't expecting.  */ 
#endif

#ifndef DRIVER_SENSE
#define DRIVER_SENSE 0x08
#endif

#define MAX_SCG     8   /* Max # of SCSI controllers */
#define MAX_TGT     16  /* Increased from 8 to 16 - dpg */
#define MAX_LUN     8

LOCAL   long sg_max_dma = -1;
LOCAL   short   scgfiles[MAX_SCG][MAX_TGT][MAX_LUN];
LOCAL   short   buscookies[MAX_SCG];

#ifdef  SG_BIG_BUFF
#define MAX_DMA_LINUX   SG_BIG_BUFF /* Defined in include/scsi/sg.h */
#else
#define MAX_DMA_LINUX   (4*1024)    /* Old Linux versions       */
#endif
#define SG_NV_MAX_DMA_LINUX   131072

LOCAL   int pack_id = 5;        /* Should be a random number    */
LOCAL   char    *SCSIbuf = (char *)-1;

#if !defined(__i386) && !defined(i386) && !defined(mc68000)
#define MISALIGN
#endif
/*#define   MISALIGN*/
/*#undef    SG_GET_BUFSIZE*/

#if defined(USE_PG) && !defined(USE_PG_ONLY)
#include "scsi-linux-pg.c"
#endif

#ifdef  MISALIGN
LOCAL   int scsi_getint __PR((int *ip));
#endif
LOCAL   int scsi_send   __PR((SCSI *scgp, int f, struct scg_cmd *sp));
LOCAL   BOOL    sg_setup    __PR((SCSI *scgp, int f, int busno, int tgt, int tlun));
LOCAL   void    sg_initdev  __PR((SCSI *scgp, int f));
LOCAL   int sg_mapbus   __PR((int ino));
LOCAL   int sg_mapbus2   __PR((int bus));
LOCAL   BOOL    sg_mapdev   __PR((SCSI *scgp, int f, int *busp, int *tgtp, int *lunp,
                            int *chanp, int *inop));
LOCAL   void    sg_settimeout   __PR((int f, int timeout));

EXPORT
int scsi_open(scgp, device, busno, tgt, tlun)
    SCSI    *scgp;      /* dpg: Is there a justification for using function */
    char    *device;    /* declarations deprecated since 1989 (ISO C) ?? */
    int busno;
    int tgt;
    int tlun;
{
    register int    f;
    register int    i;
    register int    b;
    register int    t;
    register int    l;
    register int    nopen = 0;
    char        devname[64];

    for (b=0; b < MAX_SCG; b++) {
        buscookies[b] = (short)-1;
        for (t=0; t < MAX_TGT; t++) {
            for (l=0; l < MAX_LUN ; l++)
                scgfiles[b][t][l] = (short)-1;
        }
    }
    if ((device != NULL && *device != '\0') || (busno == -2 && tgt == -2))
        goto openbydev;

    if (busno >= 0 && tgt >= 0 && tlun >= 0) {
        if (busno >= MAX_SCG || tgt >= MAX_TGT || tlun >= MAX_LUN)
            return (-1);
    }

    for (i=0; i < 32; i++) {
        sprintf(devname, "/dev/sg%d", i);
        f = open(devname, O_RDWR | O_NONBLOCK);
        if (f < 0) {
            if (errno != ENOENT && errno != ENXIO && errno != ENODEV &&
                errno != EBUSY)
                comerr("Cannot open '%s'.\n", devname);
        } else {
            if (sg_setup(scgp, f, busno, tgt, tlun))
                return (++nopen);
            if (busno < 0 && tgt < 0 && tlun < 0)
                nopen++;
        }
    }
    if (nopen == 0) for (i=0; i <= 25; i++) {
        sprintf(devname, "/dev/sg%c", i+'a');
        f = open(devname, O_RDWR | O_NONBLOCK);
        if (f < 0) {
            if (errno != ENOENT && errno != ENXIO && errno != ENODEV &&
                errno != EBUSY)
                comerr("Cannot open '%s'.\n", devname);
        } else {
            if (sg_setup(scgp, f, busno, tgt, tlun))
                return (++nopen);
            if (busno < 0 && tgt < 0 && tlun < 0)
                nopen++;
        }
    }
openbydev:
    if (device != NULL && *device != '\0') {
        f = open(device, O_RDWR | O_NONBLOCK);
        if (f < 0) {
            if (errno == ENOENT)
                goto openpg;
            else
                comerr("Cannot open '%s'.\n", device);
        }
        else {
            if (!sg_mapdev(scgp, f, &busno, &tgt, &tlun, 0, 0)) {
                close(f);
                goto openpg;
            }

            if (scgp->scsibus < 0)
                scgp->scsibus = busno;
            if (scgp->target < 0)
                scgp->target = tgt;
            if (scgp->lun < 0)
                scgp->lun = tlun;

            if (sg_setup(scgp, f, busno, tgt, tlun))
                return (++nopen);
        }
    }
openpg:
#ifdef  USE_PG
    nopen += pg_open(scgp, device, busno, tgt, tlun);
#endif
    return (nopen);
}

LOCAL BOOL
sg_setup(scgp, f, busno, tgt, tlun)
    SCSI    *scgp;
    int f;
    int busno;
    int tgt;
    int tlun;
{
    int n;
    int Chan;
    int Ino;
    int Bus;
    int Target;
    int Lun;
    BOOL    onetarget = FALSE;
    int val;


    if (SG_INTF_UNKN == sg_intf) { /* determine which sg interface we have */
        if ((val = ioctl(f, SG_GET_BUFSIZE, 0)) < 0) { 
            if (ioctl(f, SG_GET_RESERVED_SIZE, &val) < 0)
                sg_intf = SG_INTF_ORIG;
            else
                sg_intf = SG_INTF_JA;
        }
        else
            sg_intf = SG_INTF_NV;
    }
    if (scgp->scsibus >= 0 && scgp->target >= 0 && scgp->lun >= 0)
        onetarget = TRUE;

    sg_mapdev(scgp, f, &Bus, &Target, &Lun, &Chan, &Ino);
#ifdef DPG_DEBUG
    printf("sg_setup: bs=%d, sid=%d, lun=%d, intf=%d, 1trg=%d\n",
           busno, tgt, tlun, sg_intf, onetarget);
    printf(" Bs=%d, Sid=%d, Lun=%d, Chan=%d, Ino=%d\n",
           Bus, Target, Lun, Chan, Ino);
#endif

    /*
     * For old kernels try to make the best guess.
     */
    if (-1 == Bus) { /* What is happening here, I don't understand,
                        but if it's a newish kernel, it breaks */
        Ino |= Chan << 8;
        n = sg_mapbus(Ino);
    }
    else
        n = sg_mapbus2(Bus);     /* if Bus number available, use it */

    if (Bus == -1) {
        Bus = n;
        if (scgp->debug)
            printf("SCSI Bus: %d (mapped from %d)\n", Bus, Ino);
    }

    if (scgfiles[Bus][Target][Lun] == (short)-1)
        scgfiles[Bus][Target][Lun] = (short)f;

    if (onetarget) {
        if (Bus == busno && Target == tgt && Lun == tlun) {
            sg_initdev(scgp, f);
            return (TRUE);
        } else {
            scgfiles[Bus][Target][Lun] = (short)-1;
            close(f);
        }
    } else
        sg_initdev(scgp, f);
    return (FALSE);
}

LOCAL void
sg_initdev(scgp, f)
    SCSI    *scgp;
    int f;
{
    struct sg_rep {
        struct sg_header    hd;
        unsigned char       rbuf[100];
    } sg_rep;
    int n;
    long maxdma = MAX_DMA_LINUX;
    int val;

    /* Eat any unwanted garbage from prior use of this device */

    n = fcntl(f, F_GETFL);  /* Be very proper about this */
    /* fcntl(f, F_SETFL, n|O_NONBLOCK); opened non-blocking ... */

    if (SG_INTF_JA != sg_intf) { /* pretty horrible, not needed in new stuff */
        fillbytes((caddr_t)&sg_rep, sizeof(struct sg_header), '\0');
        sg_rep.hd.reply_len = sizeof(struct sg_header);
        while (read(f, &sg_rep, sizeof(sg_rep)) >= 0 || errno != EAGAIN)
            ;
    }

    fcntl(f, F_SETFL, n & (~ O_NONBLOCK));    /* turn off non-blocking */
    sg_settimeout(f, scgp->deftimeout);

    switch (sg_intf)
    {
    case SG_INTF_JA:
        if (ioctl(f, SG_GET_RESERVED_SIZE, &val) >= 0)
            maxdma = val;
        break;
    case SG_INTF_NV:
        /*
         * We assume that all /dev/sg instances use the same maximum
         * buffer size. Bad assumption - dpg
         */
        if ((maxdma = ioctl(f, SG_GET_BUFSIZE, 0)) < 0)
            maxdma = MAX_DMA_LINUX;
        break;
    default:
        break;
    }
    if (sg_max_dma < 0)
        sg_max_dma = maxdma;
    else if (maxdma < sg_max_dma)
        sg_max_dma = maxdma;
}

LOCAL int
sg_mapbus(ino)
    int ino;
{
    register int    i;

    for (i=0; i < MAX_SCG; i++) {
        if (buscookies[i] == (short)-1) {
            buscookies[i] = ino;
            return (i);
        }

        if (buscookies[i] == ino)
            return (i);
    }
    return (0);
}

LOCAL int
sg_mapbus2(bus)
    int bus;
{
    register int    k;

    if ((bus >= 0) && (bus < MAX_SCG) && (buscookies[bus] < 0)) {
        buscookies[bus] = (short)bus;
        for (k = 0; k < bus; ++k) {
            if (-1 == buscookies[k])
                buscookies[k] = (short)-2;      /* fill in holes */
        }
        return bus;
    }
    return 0;
}

LOCAL BOOL
sg_mapdev(scgp, f, busp, tgtp, lunp, chanp, inop)
    SCSI    *scgp;
    int f;
    int *busp;
    int *tgtp;
    int *lunp;
    int *chanp;
    int *inop;
{
    struct  sg_id {
        long    l1; /* target | lun << 8 | channel << 16 | low_ino << 24 */
        long    l2; /* Unique id */
    } sg_id;
    int Chan;
    int Ino;
    int Bus;
    int Target;
    int Lun;

    if (ioctl(f, SCSI_IOCTL_GET_IDLUN, &sg_id))
        return (FALSE);
    if (scgp->debug)
        printf("l1: 0x%lX l2: 0x%lX\n", sg_id.l1, sg_id.l2);
    if (ioctl(f, SCSI_IOCTL_GET_BUS_NUMBER, &Bus) < 0) {
        Bus = -1;
    }

    Target  = sg_id.l1 & 0xFF;
    Lun = (sg_id.l1 >> 8) & 0xFF;
    Chan    = (sg_id.l1 >> 16) & 0xFF;
    Ino = (sg_id.l1 >> 24) & 0xFF;
    if (scgp->debug) {
        printf("Bus: %d Target: %d Lun: %d Chan: %d Ino: %d\n",
                Bus, Target, Lun, Chan, Ino);
    }
    *busp = Bus;
    *tgtp = Target;
    *lunp = Lun;
    if (chanp)
        *chanp = Chan;
    if (inop)
        *inop = Ino;
    return (TRUE);
}

LOCAL long
scsi_maxdma(scgp)
    SCSI    *scgp;
{
    long maxdma = MAX_DMA_LINUX;

    if (sg_max_dma >= 0)
        maxdma = sg_max_dma;
    /* N.B. With SG_INTF_JA a maxdma of 0 is possible and in this case
       cdrecord should not attempt to burn a cd since there is a very
       high probability of ENOMEM during a write() to the sg device.
       [Other variants of the sg interface face the same problem but
        would have crashed already or misinform the user concerning
        this "out-of-resource" situation.] -dpg  */

#ifdef  USE_PG
    if (scgp->scsibus == pgbus)
        return (pg_maxdma(scgp));
    if ((scgp->scsibus < 0) && (pg_maxdma(scgp) < maxdma))
        return (pg_maxdma(scgp));
#endif
    return maxdma;
}

EXPORT void *
scsi_getbuf(scgp, amt)
    SCSI    *scgp;
    long    amt;
{
    char    *ret;
    long maxdma;

    if (scg_maxdma == 0)
        scg_maxdma = scsi_maxdma(scgp);

    if (amt <= 0 || amt > scg_maxdma)
        return ((void *)0);
    if (scgp->debug)
        printf("scsi_getbuf: %ld bytes\n", amt);
    /*
     * For performance reason, we allocate pagesize()
     * bytes before the SCSI buffer to avoid
     * copying the whole buffer contents when
     * setting up the /dev/sg data structures.
     */
    ret = valloc((size_t)(amt+getpagesize()));
    if (ret == NULL)
        return (ret);
    ret += getpagesize();
    SCSIbuf = ret;
    return ((void *)ret);
}

EXPORT
BOOL scsi_havebus(scgp, busno)
    SCSI    *scgp;
    int busno;
{
    register int    t;
    register int    l;

    if (busno < 0 || busno >= MAX_SCG)
        return (FALSE);

    for (t=0; t < MAX_TGT; t++) {
        for (l=0; l < MAX_LUN ; l++)
            if (scgfiles[busno][t][l] >= 0)
                return (TRUE);
    }
    return (FALSE);
}

EXPORT
int scsi_fileno(scgp, busno, tgt, tlun)
    SCSI    *scgp;
    int busno;
    int tgt;
    int tlun;
{
    if (busno < 0 || busno >= MAX_SCG ||
        tgt < 0 || tgt >= MAX_TGT ||
        tlun < 0 || tlun >= MAX_LUN)
        return (-1);

    return ((int)scgfiles[busno][tgt][tlun]);
}

EXPORT
int scsi_isatapi(scgp)
    SCSI    *scgp;
{
#ifdef  USE_PG
    if (scgp->scsibus == pgbus)
        return (pg_isatapi(scgp));
#endif

#ifdef  SG_EMULATED_HOST
    {
    int emulated = FALSE;
    int f = scsi_fileno(scgp, scgp->scsibus, scgp->target, scgp->lun);

    if (ioctl(f, SG_EMULATED_HOST, &emulated) >= 0)
        return (emulated != 0);
    }
#endif
    return (-1);
}

EXPORT
int scsireset(scgp)
    SCSI    *scgp;
{
#ifdef  USE_PG
    if (scgp->scsibus == pgbus)
        return (pg_reset(scgp));
#endif
    /*
     * Do we have a SCSI reset in the Linux sg driver? Nope - dpg
     */
    return (-1);
}

LOCAL void
sg_settimeout(f, tmo)
    int f;
    int tmo;
{
    tmo *= HZ;
    if (tmo)
        tmo += HZ/2;

#ifdef DPG_DEBUG
    /* tmo = 0; */    /* try and simulate time out errors */
#endif
    if (ioctl(f, SG_SET_TIMEOUT, &tmo) < 0)
        comerr("Cannot set SG_SET_TIMEOUT.\n");
}

/*
 * Get misaligned int.
 * Needed for all recent processors (sparc/ppc/alpha)
 * because the /dev/sg design forces us to do misaligned
 * reads of integers.
 */
#ifdef  MISALIGN
LOCAL int
scsi_getint(ip)
    int *ip;
{
         int    ret;
    register char   *cp = (char *)ip;
    register char   *tp = (char *)&ret;
    register int    i;

    for (i = sizeof(int); --i >= 0; )
        *tp++ = *cp++;
    
    return (ret);
}
#define GETINT(a)   scsi_getint(&(a))
#else
#define GETINT(a)   (a)
#endif

LOCAL int
scsi_send(scgp, f, sp)
    SCSI        *scgp;
    int     f;
    struct scg_cmd  *sp;
{
    struct sg_rq    *sgp;
    struct sg_rq    *sgp2;
    int i;
    int pack_len;
    int reply_len;
    int amt = sp->cdb_len;
    struct sg_rq {
        struct sg_header    hd;
        unsigned char       buf[SG_NV_MAX_DMA_LINUX+SCG_MAX_CMD];
    } sg_rq;
    struct sg_rq_nv_t {
        Sgh_nv    hd;
    } * sgh_nvp;
    struct sg_rq_ja_t {
        Sgh_ja    hd;
    } * sgh_jap;
    char    driver_byte;
    char    host_byte;
    char    msg_byte;
    char    status_byte;
    int     trg_status;

    if (f < 0) {
        sp->error = SCG_FATAL;
        return (0);
    }
#ifdef  USE_PG
#ifdef DPG_DEBUG
    printf("scsi_send: scgp->scsibus=%d pgbus=%d\n", scgp->scsibus, pgbus);
#endif
    if (scgp->scsibus == pgbus)
        return (pg_send(scgp, f, sp));
#endif
    if (sp->timeout != scgp->deftimeout)
        sg_settimeout(f, sp->timeout);


    sgp2 = sgp = &sg_rq;
    if (sp->addr == SCSIbuf) {
        sgp = (struct sg_rq *)
            (SCSIbuf - (sizeof(struct sg_header) + amt));
        sgp2 = (struct sg_rq *)
            (SCSIbuf - (sizeof(struct sg_header)));
    } else {
        if (scgp->debug) {
            printf("DMA addr: 0x%8.8lX size: %d - using copy buffer\n",
                (long)sp->addr, sp->size);
        }
        if (sp->size > (int)(sizeof(sg_rq.buf) - SCG_MAX_CMD)) {
            errno = ENOMEM;
            return (-1);
        }
    }

    /*
     * This is done to avoid misaligned access of sgp->some_int
     */
    pack_len = sizeof(struct sg_header) + amt;
    reply_len = sizeof(struct sg_header);
    if (sp->flags & SCG_RECV_DATA) {
        reply_len += sp->size;
    } else {
        pack_len += sp->size;
    }

#ifdef  MISALIGN
    /*
     * sgp->some_int may be misaligned if (sp->addr == SCSIbuf)
     * This is no problem on Intel porocessors, however
     * all other processors don't like it.
     * sizeof(struct sg_header) + amt is usually not a multiple of
     * sizeof(int). For this reason, we fill in the values into sg_rq
     * which is always corectly aligned and then copy it to the real
     * location if this location differs from sg_rq.
     * Never read/write directly to sgp->some_int !!!!!
     */
    fillbytes((caddr_t)&sg_rq, sizeof(struct sg_header), '\0');

    sg_rq.hd.pack_len = pack_len;
    sg_rq.hd.reply_len = reply_len;
    sg_rq.hd.pack_id = pack_id++;
/*  sg_rq.hd.result = 0;    not needed because of fillbytes() */

    if ((caddr_t)&sg_rq != (caddr_t)sgp)
        movebytes((caddr_t)&sg_rq, (caddr_t)sgp, sizeof(struct sg_header));
#else
    fillbytes((caddr_t)sgp, sizeof(struct sg_header), '\0');

    sgp->hd.pack_len = pack_len;
    sgp->hd.reply_len = reply_len;
    sgp->hd.pack_id = pack_id++;
/*  sgp->hd.result = 0; not needed because of fillbytes() */
#endif
    if (amt == 12)
        sgp->hd.twelve_byte = 1;


    for (i = 0; i < amt; i++ ) {
        sgp->buf[i] = sp->cdb.cmd_cdb[i];;
    }
    if (!(sp->flags & SCG_RECV_DATA)) {
        if ((void *)sp->addr != (void *)&sgp->buf[amt])
            movebytes(sp->addr, &sgp->buf[amt], sp->size);
        amt += sp->size;
    }
    if (SG_INTF_NV == sg_intf) {
        sgh_nvp = (struct sg_rq_nv_t *)sgp;
        sgh_nvp->hd.want_new  = 1;          /* Order new behaviour  */
        sgh_nvp->hd.cdb_len   = sp->cdb_len;    /* Set CDB length   */
        if (sp->sense_len > SG_MAX_SENSE)
            sgh_nvp->hd.sense_len = SG_MAX_SENSE;
        else
            sgh_nvp->hd.sense_len = sp->sense_len;
    }
    i = sizeof(struct sg_header) + amt;
    if ((amt = write(f, sgp, i)) < 0) {         /* write */
        sg_settimeout(f, scgp->deftimeout);
        return (-1);
    } else if (amt != i) {
        errmsg("scsi_send(%s) wrote %d bytes (expected %d).\n",
                        scsi_command, amt, i);
    }

    if (sp->addr == SCSIbuf) {
        movebytes(sgp, sgp2, sizeof(struct sg_header));
        sgp = sgp2;
    }
    sgp->hd.sense_buffer[0] = 0;
    if ((amt = read(f, sgp, reply_len)) < 0) {      /* read */
        sg_settimeout(f, scgp->deftimeout);
        return (-1);
    }

    if (sp->flags & SCG_RECV_DATA && ((void *)sgp->buf != (void *)sp->addr)) {
        movebytes(sgp->buf, sp->addr, sp->size);
    }
    sp->ux_errno = GETINT(sgp->hd.result);      /* Unaligned read */
    sp->error = SCG_NO_ERROR;

    sgh_nvp = (struct sg_rq_nv_t *)sgp;
    if (((SG_INTF_NV == sg_intf) && sgh_nvp->hd.grant_new) || 
        (SG_INTF_JA == sg_intf)) {
        if (SG_INTF_NV == sg_intf) {
            sp->sense_count = sgh_nvp->hd.sense_len;
            pack_len = GETINT(sgh_nvp->hd.sg_cmd_status); /* Unaligned read */
            driver_byte = (pack_len  >> 24) & 0xFF;
            host_byte   = (pack_len  >> 16) & 0xFF;
            msg_byte    = (pack_len  >> 8) & 0xFF;
            status_byte =  pack_len  & 0xFF;
/* N.B. Linux's status_byte is _not_ the same as the SCSI standard's target
   target, but has proprietary bits masked off and is shifted right 1 bit! */
        }
        else {
            sgh_jap = (struct sg_rq_ja_t *)sgp;
            sp->sense_count = SG_MAX_SENSE;
            trg_status = sgh_jap->hd.target_status;
            status_byte = (trg_status << 1) & 0x3e;
            host_byte = sgh_jap->hd.host_status;
            driver_byte = sgh_jap->hd.driver_status;
            msg_byte = 0;
        }

        switch (host_byte) {

        case DID_OK:
                if (SG_INTF_NV == sg_intf) {
                    if (sgp->hd.sense_buffer[0] != 0) {
                    /*
                     * The Linux SCSI system up to 2.1.xx
                     * trashes the status byte in the
                     * kernel. Until this gets fixed, we
                     * need this hack.
                     * This feature seems to be associated with the scsi
                     * ide emulation that returns a drive_status byte 
                     * masked with DRIVER_SENSE when it sets the sense
                     * buffer (while status_byte is 0). -dpg
                     */
                        sp->error = SCG_RETRYABLE;
                        status_byte = SG_SCSI_CHECK_CONDITION;
                        /* sgp->hd.sense_len = SG_MAX_SENSE; */
                    }
                }
                else if (DRIVER_SENSE & driver_byte) {
                    sp->error = SCG_RETRYABLE;
                    if (0 == status_byte)
                        status_byte = SG_SCSI_CHECK_CONDITION;
                    /* sgp->hd.sense_len = SG_MAX_SENSE; */
                }
                break;
        case DID_NO_CONNECT:    /* Arbitration won, retry NO_CONNECT? */
        case DID_BAD_TARGET:
                sp->error = SCG_FATAL;
                break;
    
        case DID_TIME_OUT:
                sp->error = SCG_TIMEOUT;
                break;

        default:
                sp->error = SCG_RETRYABLE;
                break;

        }
        if ((host_byte != DID_OK || status_byte != 0) && sp->ux_errno == 0)
            sp->ux_errno = EIO;
        sp->u_scb.cmd_scb[0] = status_byte;
        if (status_byte & SG_SCSI_CHECK_CONDITION) {
            /* Picks up CHECK_CONDITION and COMMAND_TERMINATED. */
            if (SG_INTF_NV == sg_intf)
                sp->sense_count = sgh_nvp->hd.sense_len;
            else
                sp->sense_count = SG_MAX_SENSE;
            movebytes(sgp->hd.sense_buffer, sp->u_sense.cmd_sense, sp->sense_count);
        }
    } else {
        if (GETINT(sgp->hd.result) == EBUSY) {  /* Unaligned read */
            struct timeval to;

            to.tv_sec = sp->timeout;
            to.tv_usec = 500000;
            scsitimes(scgp);

            if (scgp->cmdstop->tv_sec < to.tv_sec ||
                (scgp->cmdstop->tv_sec == to.tv_sec &&
                scgp->cmdstop->tv_usec < to.tv_usec)) {

                sp->ux_errno = 0;
                sp->error = SCG_TIMEOUT;    /* a timeout */
            } else {
                sp->error = SCG_RETRYABLE;  /* may be BUS_BUSY */
            }
        }

        if (sp->flags & SCG_RECV_DATA)
            sp->resid = (sp->size + sizeof(struct sg_header)) - amt;
        else
            sp->resid = 0;  /* sg version1 cannot return DMA resid count */

        if (sgp->hd.sense_buffer[0] != 0) {
            sp->error = SCG_RETRYABLE;
            sp->scb.chk = 1;
            sp->sense_count = SG_MAX_SENSE;
            movebytes(sgp->hd.sense_buffer, sp->u_sense.cmd_sense, sp->sense_count);
            if (sp->ux_errno == 0)
                sp->ux_errno = EIO;
        }
    }

    if (scgp->verbose > 0 && scgp->debug) {
        switch (sg_intf)
        {
        case SG_INTF_NV:
            sgh_nvp = (struct sg_rq_nv_t *)sgp;
            printf("status: %X pack_len: %d, reply_len: %d pack_id: %d result: %d wn: %d gn: %d cdb_len: %d sense_len: %d sense[0]: %02X\n",
                GETINT(sgh_nvp->hd.sg_cmd_status),
                GETINT(sgh_nvp->hd.pack_len),
                GETINT(sgh_nvp->hd.reply_len),
                GETINT(sgh_nvp->hd.pack_id),
                GETINT(sgh_nvp->hd.result),
                sgh_nvp->hd.want_new,
                sgh_nvp->hd.grant_new,
                sgh_nvp->hd.cdb_len,
                sgh_nvp->hd.sense_len,  
                sgp->hd.sense_buffer[0]);
            break;
        case SG_INTF_JA:
            printf("pack_len: %d, reply_len: %d pack_id: %d result: %d trg_s: %x h_s: %x d_s: %x sense[0]: %02X\n",
                GETINT(sgp->hd.pack_len),
                GETINT(sgp->hd.reply_len),
                GETINT(sgp->hd.pack_id),
                GETINT(sgp->hd.result),
                trg_status,
                (int)host_byte,
                (int)driver_byte,
                (int)sgh_jap->hd.sense_buffer[0]);
            break;
        default:
            printf("pack_len: %d, reply_len: %d pack_id: %d result: %d sense[0]: %02X\n",
                    GETINT(sgp->hd.pack_len),
                    GETINT(sgp->hd.reply_len),
                    GETINT(sgp->hd.pack_id),
                    GETINT(sgp->hd.result),
                    sgp->hd.sense_buffer[0]);
        }
#ifdef  DEBUG
        printf("sense: ");
        for (i=0; i < 16; i++)
            printf("%02X ", sgp->hd.sense_buffer[i]);
        printf("\n");
#endif
    }

    if (sp->timeout != scgp->deftimeout)
        sg_settimeout(f, scgp->deftimeout);
    return 0;
}
